"""Module with search interface."""
from typing import Dict, Iterable, Optional

from jira import JIRA, Issue
from jira.client import ResultList

from jirapt.jira_response_iterable import JiraResponseIterable, JiraResponseIterableParallel


def search_issues(
    jira: JIRA, jql: str, n_threads: Optional[int] = None, **kwargs: Optional[Dict]
) -> Iterable[Issue]:
    """
    Lazily search for paginated jira issues.

    :param jira: Jira API instance.
    :param jql: JQL query to search.
    :param n_threads: Number of threads to use for execution.
    :param kwargs: Additional search_issues arguments.
    :return: Iterable of Jira issues found.
    """

    def get_more_fn(start_idx: int) -> ResultList:
        return jira.search_issues(jql, startAt=start_idx, **kwargs)

    if n_threads is not None:
        return JiraResponseIterableParallel(get_more_fn(0), get_more_fn, n_threads)
    return JiraResponseIterable(get_more_fn(0), get_more_fn)



"""Iterable for jira requests."""
import math
from concurrent.futures import ThreadPoolExecutor as Executor
from queue import Queue
from typing import Callable, Iterator

from jira.client import ResultList

DEFAULT_WORKERS = 4


class JiraResponseIterable(object):
    """Object to iterate over paginated jira results."""

    def __init__(self, results: ResultList, get_more_fn: Callable[[int], ResultList],) -> None:
        """
        Create a JiraResponseIterable.

        :param results: Response object from Jira.
        :param get_more_fn: Function to get next page of data.
        """
        self.results = results
        self.get_more_fn = get_more_fn

    def __iter__(self) -> Iterator:
        """Get next item from jira."""
        while True:
            for item in self.results:
                yield item

            # Have we looked at all the results?
            if self.results.startAt + self.results.maxResults > self.results.total:
                break

            self.results = self.get_more_fn(self.results.startAt + self.results.maxResults)

    def __len__(self) -> int:
        """Get the total number of items available."""
        return self.results.total


class JiraResponseIterableParallel(object):
    """
    Create a response iterator that will gather results in multiple threads.

    Note: This will gather all results in memory at the start of iteration. So it may use a lot
    of memory. It will also wait for all results to be collected before starting iteration.
    """

    def __init__(
        self,
        results: ResultList,
        get_more_fn: Callable[[int], ResultList],
        n_workers: int = DEFAULT_WORKERS,
    ) -> None:
        """
        Create a JiraResponseIterableParallel.

        :param results: Response object from Jira.
        :param get_more_fn: Function to get next page of data.
        :param n_workers: Number of threads to use for processing.
        """
        self.page_size = results.maxResults
        self.total_results = results.total
        self.get_more_fn = get_more_fn
        self.n_workers = n_workers
        self.queue: Queue = Queue()

        self._add_results_to_queue(results)

    def _add_results_to_queue(self, results: ResultList) -> None:
        """
        Add a set of jira responses to the queue.

        :param results: Results to add to queue.
        """
        for r in results:
            self.queue.put(r)

    def _get_more(self, index: int) -> ResultList:
        """
        Get the specified page of results from jira.

        :param index: Index of page to query.
        :return:
        """
        return self.get_more_fn(index * self.page_size)

    def _gather_results(self) -> None:
        """Spawn workers and gather results from jira."""
        total_workers = math.ceil(self.total_results / self.page_size)
        with Executor(max_workers=self.n_workers) as exe:
            workers = [exe.submit(self._get_more, i) for i in range(1, total_workers)]

            for w in workers:
                result = w.result()
                self._add_results_to_queue(result)

    def __iter__(self) -> Iterator:
        """Get next item from jira."""
        self._gather_results()
        i = 0
        while i < self.total_results:
            item = self.queue.get()
            yield item
            i += 1

    def __len__(self) -> int:
        """Get the total number of items available."""
        return self.total_results



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jirapt']

package_data = \
{'': ['*']}

install_requires = \
['jira>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'jira-power-tools',
    'version': '0.2.1',
    'description': 'Tools to improve working with the python jira API.',
    'long_description': '# jira power tools\n\nA collection of tools to make it easier to work with the \n[Python Jira API](https://pypi.org/project/jira/).\n\n[![PyPI](https://img.shields.io/pypi/v/jira-power-tools.svg)](https://pypi.org/project/jira-power-tools/) ![Build](https://github.com/dbradf/jira-power-tools/workflows/Test%20Python%20Package/badge.svg)\n\n## Usages\n\n### Lazy Pagination of Jira Issues\n\nIf you are querying a lot of jira issues. Paginating the results can lighten the memory usage\non the Jira server. But needing to keep track of the pagination is painful. Using lazy pagination\ngives you an iterable that can track pagination for you:\n\n```python\nimport jirapt\n\njira = # jira server instance.\njql = "JQL query"\n\nissues = jirapt.search_issues(jira, jql, ...) # you can include any parameters you might pass to search_issues.\n\nfor issue in issues:\n    # perform work on issue\n```\n\nNote: You gather results in parallel by specifying `n_threads=N`. This will gather all the results\nbefore starting iteration. So there will be a delay before the first iteration starts and all the\nresults will need to fit into memory.\n',
    'author': 'David Bradford',
    'author_email': 'david.bradford@mongodb.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/...',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
